@@ -4,7 +4,7 @@ Created on 2023-11-06
 @author: wf
 """
 from ngwidgets.cmd import WebserverCmd
-from dcm.dcm_webserver import DynamiceCompentenceMapWebServer
+from dcm.dcm_webserver import DynamicCompentenceMapWebServer
 import sys
 from dcm.dcm_core import DynamicCompetenceMap
 from argparse import ArgumentParser
@@ -20,14 +20,14 @@ class CompetenceCmd(WebserverCmd):
         """        
         parser=super().getArgParser(description, version_msg)
         parser.add_argument("-v", "--verbose", action="store_true", help="show verbose output [default: %(default)s]")
-        parser.add_argument("-rp", "--root_path",default=DynamicCompetenceMap.examples_path(),help="path to example json files [default: %(default)s]")
+        parser.add_argument("-rp", "--root_path",default=DynamicCompetenceMap.examples_path(),help="path to example dcm definition files [default: %(default)s]")
         return parser
     
 def main(argv:list=None):
     """
     main call
     """
-    cmd=CompetenceCmd(config=DynamiceCompentenceMapWebServer.get_config(),webserver_cls=DynamiceCompentenceMapWebServer)
+    cmd=CompetenceCmd(config=DynamicCompentenceMapWebServer.get_config(),webserver_cls=DynamicCompentenceMapWebServer)
     exit_code=cmd.cmd_main(argv)
     return exit_code
         
